# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# xoutil.inspect
# ----------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# All rights reserved
#
# This file is distributed under the terms of the LICENCE distributed
# with this package.
#
# Created 2014-05-02


'''Extensions to Python's ``inspect`` module.

You may use it as drop-in replacement of ``inspect``.  Although we don't
document all items here.  Refer to :mod:`inspect's <inspect>` documentation.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _absolute_import)


from xoutil.modules import copy_members as _copy_python_module_members
_pm = _copy_python_module_members()
types = _pm.types

isdatadescriptor = _pm.isdatadescriptor

# TODO: Rename all use of `StandardException`
from xoutil.eight.exceptions import StandardError as StandardException

try:
    getattr_static = _pm.getattr_static
except AttributeError:
    # Copied from `/usr/lib/python3.3/inspect.py`
    # TODO: Maybe all these function must be moved to eight
    from xoutil.eight import typeof as _typeof
    _sentinel = object()

    def _safe_search_bases(cls, accum):
        # Simulate the "classic class" search order.
        if cls in accum:
            return
        else:

            accum.append(cls)
            for base in cls.__bases__:
                _safe_search_bases(base, accum)

    def _static_getmro(klass):
        '''Get a reasonable method resolution order of a class.

        Works well for both old-style and new-style classes.

        '''
        try:
            old_class_type = types.ClassType
        except AttributeError:
            # Python 3.1 lacks both getattr_static and ClassType
            class ClassType(type):
                '''Impossible class type.'''
            old_class_type = ClassType
        if isinstance(klass, type):
            return type.__dict__['__mro__'].__get__(klass)
        elif isinstance(klass, old_class_type):
            res = []
            _safe_search_bases(klass, res)
            return res
        else:
            msg = "doesn't apply to '%s' object"
            raise TypeError(msg % _typeof(klass).__name__)

    def _check_instance(obj, attr):
        try:
            instance_dict = object.__getattribute__(obj, "__dict__")
        except AttributeError:
            instance_dict = {}
        return dict.get(instance_dict, attr, _sentinel)

    def _check_class(klass, attr):
        for entry in _static_getmro(klass):
            if _shadowed_dict(type(entry)) is _sentinel:
                try:
                    return entry.__dict__[attr]
                except KeyError:
                    pass
        return _sentinel

    def _is_type(obj):
        try:
            _static_getmro(obj)
            return True
        except StandardException:
            return False

    def _shadowed_dict(klass):
        if isinstance(klass, type):
            dict_get = type.__dict__["__dict__"].__get__
        else:
            dict_get = lambda item: {'__dict__': item.__dict__}
        for entry in _static_getmro(klass):
            try:
                class_dict = dict_get(entry)["__dict__"]
            except KeyError:
                pass
            else:
                _desc = isinstance(class_dict, types.GetSetDescriptorType)
                if (not _desc and class_dict.__name__ == "__dict__" and
                        class_dict.__objclass__ is entry):
                    return class_dict
        return _sentinel

    def _is_descriptor(klass_result):
        _ktype = type(klass_result)
        return (_check_class(_ktype, '__get__') is not _sentinel and
                _check_class(_ktype, '__set__') is not _sentinel)

    def getattr_static(obj, attr, default=_sentinel):
        """Retrieve attributes without triggering dynamic lookup via the
           descriptor protocol,  __getattr__ or __getattribute__.

           Note: this function may not be able to retrieve all attributes
           that getattr can fetch (like dynamically created attributes)
           and may find attributes that getattr can't (like descriptors
           that raise AttributeError). It can also return descriptor objects
           instead of instance members in some cases.  See the
           documentation for details.
        """
        instance_result = _sentinel
        if not _is_type(obj):
            _member = types.MemberDescriptorType
            klass = _typeof(obj)
            dict_attr = _shadowed_dict(klass)
            if dict_attr is _sentinel or isinstance(dict_attr, _member):
                instance_result = _check_instance(obj, attr)
        else:
            klass = obj

        klass_result = _check_class(klass, attr)

        if (instance_result is not _sentinel and
                klass_result is not _sentinel and
                _is_descriptor(klass_result)):
            return klass_result
        elif instance_result is not _sentinel:
            return instance_result
        elif klass_result is not _sentinel:
            return klass_result
        if obj is klass:
            if isinstance(obj, type):
                # for types we check the metaclass too
                meta_result = _check_class(type(klass), attr)
                if meta_result is not _sentinel:
                    return meta_result
            elif attr == '__name__':
                try:
                    return klass.__name__
                except AttributeError:
                    pass
        if default is not _sentinel:
            return default
        else:
            print('>>>', default)
            raise AttributeError(attr)


def get_attr_value(obj, name, *default):
    '''Get a named attribute from an object in a safe way.

    Similar to `getattr` but without triggering dynamic look-up via the
    descriptor protocol, `__getattr__` or `__getattribute__` by using
    `getattr_static`:func:.

    '''
    from xoutil import Undefined as _undef
    from xoutil.tools import get_default
    default = get_default(default, _undef)
    is_type = isinstance(obj, type)
    res = getattr_static(obj, name, _undef)
    if isdatadescriptor(res):
        try:
            owner = type if is_type else type(obj)
            res = res.__get__(obj, owner)
        except BaseException:
            res = _undef
    if res is _undef and not is_type:
        cls = type(obj)
        res = getattr_static(cls, name, _undef)
        if isdatadescriptor(res):
            try:
                res = res.__get__(obj, cls)
            except StandardException:
                try:
                    res = res.__get__(cls, type)
                except StandardException:
                    res = _undef
    if res is not _undef:
        return res
    elif default is not _undef:
        return default
    else:
        from xoutil.eight import typeof
        msg = "'%s' object has no attribute '%s'"
        raise AttributeError(msg % (typeof(obj).__name__, name))


def type_name(obj, affirm=False):
    '''Return the internal name for a type or a callable.

    This function is safe.  If :param obj: is not an instance of a proper type
    then returns the following depending on :param affirm:

    - If False returns None.

    - If True convert a single object to its type before returns the name, but
      if is a tuple, list or set; returns a string with a representation of
      contained types.

    Examples::

      >>> type_name(int)
      'int'

      >>> type_name(0) is None
      True

      >>> type_name(0, affirm=True)
      'int'

      >>> type_name((0, 1.1)) is None
      True

      >>> type_name((0, 1.1), affirm=True)
      '(int, float)'

    '''
    from xoutil.eight import class_types, string_types
    named_types = class_types + (types.FunctionType, types.MethodType)
    name = '__name__'
    if isinstance(obj, (staticmethod, classmethod)):
        fn = get_attr_value(obj, '__func__', None)
        if fn:
            obj = fn
    if isinstance(obj, named_types):
        # TODO: Why not use directly `get_attr_value``
        try:
            res = getattr_static(obj, name, None)
            if res:
                if isdatadescriptor(res):
                    res = res.__get__(obj, type)
        except BaseException:
            res = None
        if res is None:
            try:
                res = obj.__name__
            except AttributeError:
                res = None
    else:
        res = None
    if res is None:
        # TODO: Why not use directly `get_attr_value``
        # FIX: Improve and standardize the combination of next code
        res = getattr_static(obj, name, None)
        if res and isdatadescriptor(res):
            res = res.__get__(obj, type(obj))
    if isinstance(res, string_types):
        return res
    elif affirm:
        if isinstance(obj, (tuple, list, set)):
            if isinstance(obj, tuple):
                head, tail = '()'
            elif isinstance(obj, list):
                head, tail = '[]'
            else:
                head, tail = '{}'
            items = ', '.join(type_name(t, affirm) for t in obj)
            return str('%s%s%s' % (head, items, tail))
        else:
            return type_name(type(obj))
    else:
        return None


# TODO: Implement a safe version for `attrgetter`

if not getattr(_pm, 'getfullargspec', None):
    from xoutil.collections import namedtuple
    FullArgSpec = namedtuple(
        'FullArgSpec',
        'args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults,'
        'annotations'
    )

    def getfullargspec(func):
        import inspect
        spec = inspect.getargspec(func)
        return FullArgSpec(
            spec.args, spec.varargs, spec.keywords, spec.defaults,
            None, None, None
        )


# get rid of unused global variables
del _pm, _copy_python_module_members



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.eight.datetime.py
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2015-02-26

'''Xoutil datetime extensions for writing code that runs on Python 2 and 3

Method strftime Python 2

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)


from datetime import *    # noqa
from datetime import __doc__    # noqa


# Correctly substitute for %z and %Z escapes in strftime formats.
def _wrap_strftime(object, format, timetuple):
    # from datetime import timedelta
    import time as _time
    # Don't call utcoffset() or tzname() unless actually needed.
    freplace = None    # the string to use for %f
    zreplace = None    # the string to use for %z
    Zreplace = None    # the string to use for %Z

    # Scan format for %z and %Z escapes, replacing as needed.
    newformat = []
    push = newformat.append
    i, n = 0, len(format)
    while i < n:
        ch = format[i]
        i += 1
        if ch == '%':
            if i < n:
                ch = format[i]
                i += 1
                if ch == 'f':
                    if freplace is None:
                        freplace = '%06d' % getattr(object,
                                                    'microsecond', 0)
                    newformat.append(freplace)
                elif ch == 'z':
                    if zreplace is None:
                        zreplace = ""
                        if hasattr(object, "utcoffset"):
                            offset = object.utcoffset()
                            if offset is not None:
                                sign = '+'
                                if offset.days < 0:
                                    offset = -offset
                                    sign = '-'
                                h, m = divmod(offset, timedelta(hours=1))
                                assert not m % timedelta(minutes=1), "whole minute"
                                m //= timedelta(minutes=1)
                                zreplace = '%c%02d%02d' % (sign, h, m)
                    assert '%' not in zreplace
                    newformat.append(zreplace)
                elif ch == 'Z':
                    if Zreplace is None:
                        Zreplace = ""
                        if hasattr(object, "tzname"):
                            s = object.tzname()
                            if s is not None:
                                # strftime is going to have at this: escape %
                                Zreplace = s.replace('%', '%%')
                    newformat.append(Zreplace)
                else:
                    push('%')
                    push(ch)
            else:
                push('%')
        else:
            push(ch)
    newformat = "".join(newformat)
    print(newformat, timetuple)
    return _time.strftime(newformat, timetuple)


# def strftime(self, fmt):    # Method for class date
#     "Format using strftime()."
#     return _wrap_strftime(self, fmt, self.timetuple())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ---------------------------------------------------------------------
# xoutil.values.simple
# ---------------------------------------------------------------------
# Copyright (c) 2015 Merchise and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2015-08-26

'''Simple or internal coercers.

With coercers defined in this module, many of the `xoutil.string` utilities
could be deprecated.

In Python 3, all arrays, not only those containing valid byte or unicode
chars, are buffers.


'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from . import coercer


@coercer
def name_coerce(arg):
    '''If `arg` is a named object, return its name, else `Invalid`.

    Object names are always of `str` type, other types are considered
    invalids.

    Generator objects has the special `__name__` attribute, but are ignored
    and considered invalids.

    '''
    from types import GeneratorType
    from . import Invalid
    if isinstance(arg, GeneratorType):
        return Invalid
    else:
        if isinstance(arg, (staticmethod, classmethod)):
            fn = getattr(arg, '__func__', None)
            if fn:
                arg = fn
        res = getattr(arg, '__name__', None)
        return res if isinstance(res, str) else Invalid


@coercer
def strict_iterable_coerce(arg):
    '''Return the same argument if it is a strict iterable.

    Strings are excluded.

    '''
    from xoutil.eight import string_types
    from collections import Iterable
    from . import Invalid
    ok = not isinstance(arg, string_types) and isinstance(arg, Iterable)
    return arg if ok else Invalid


@coercer
def decode_coerce(arg):
    '''Decode objects implementing the buffer protocol.'''
    import locale
    from . import Invalid
    from xoutil.eight import text_type, callable
    encoding = locale.getpreferredencoding() or 'UTF-8'
    decode = getattr(arg, 'decode', None)
    if callable(decode):
        try:
            res = decode(encoding, 'replace')
            if not isinstance(res, text_type):
                res = None
        except BaseException:
            res = None
    else:
        res = None
    if res is None:
        try:
            # TODO: All arrays are decoded, and not only those containing
            # valid byte or unicode characters.
            import codecs
            res = codecs.decode(arg, encoding, 'replace')
        except BaseException:
            res = Invalid
    return res


@coercer
def encode_coerce(arg):
    '''Encode string objects.'''
    import locale
    from . import Invalid
    from xoutil.eight import callable
    encoding = locale.getpreferredencoding() or 'UTF-8'
    encode = getattr(arg, 'encode', None)
    if callable(encode):
        try:
            res = encode(encoding, 'replace')
            if not isinstance(res, bytes):
                res = None
        except BaseException:
            res = None
    else:
        res = None
    if res is None:
        try:
            import codecs
            res = codecs.encode(arg, encoding, 'replace')
        except BaseException:
            res = Invalid
    return res


@coercer
def unicode_coerce(arg):
    '''Decode a buffer or any object returning unicode text.

    Uses the defined `encoding` system value.

    In Python 2.x unicode has a special type different to `str` but in Python
    3 coincide with `str` type.

    Name is used in named objects, see `name_coerce`:func: for more
    information.

    See `str_coerce`:func: to coerce to standard string type, `bytes` in
    Python 2.x and unicode (`str`) in Python 3.

    .. versionadded:: 1.7.0

    '''
    from array import array
    from . import Invalid
    from xoutil.eight import text_type
    aux = name_coerce(arg)
    if aux is not Invalid:
        arg = aux
    if isinstance(arg, text_type):
        return arg
    elif isinstance(arg, bytearray):
        arg = bytes(arg)
    elif isinstance(arg, memoryview):
        arg = arg.tobytes()
    elif isinstance(arg, array):
        try:
            return arg.tounicode()
        except BaseException:
            try:
                arg = bytes(bytearray(arg.tolist()))
            except BaseException:
                arg = str(arg)
                if str is text_type:
                    return arg

    res = decode_coerce(arg)
    return text_type(arg) if res is Invalid else res


@coercer
def bytes_coerce(arg):
    '''Encode an unicode string (or any object) returning a bytes buffer.

    Uses the defined `encoding` system value.

    In Python 2.x `bytes` coincide with `str` type, in Python 3 `str` uses
    unicode and `str` is different to `bytes`.

    There are differences if you want to obtain a buffer in Python 2.x and
    Python 3; for example, the following code obtain different results::

      >>> ba = bytes([65, 66, 67])

    In Python 2.x is obtained the string ``"[65, 66, 67]"`` and in Python 3
    ``b"ABC"``.  This function normalize these differences.

    Name is used in named objects, see `name_coerce`:func: for more
    information.

    See `str_coerce`:func: to coerce to standard string type, `bytes` in
    Python 2.x and unicode (`str`) in Python 3.

    Always returns the `bytes` type.

    .. versionadded:: 1.7.0

    '''
    from array import array
    from . import Invalid
    from xoutil.eight import text_type
    aux = name_coerce(arg)
    if aux is not Invalid:
        arg = aux
    if isinstance(arg, bytes):
        return arg
    elif isinstance(arg, bytearray):
        return bytes(arg)
    elif isinstance(arg, memoryview):
        return arg.tobytes()
    elif isinstance(arg, array):
        try:
            arg = arg.tounicode()
        except BaseException:
            try:
                return bytes(bytearray(arg.tolist()))
            except BaseException:
                arg = text_type(arg)
    res = encode_coerce(arg)
    return encode_coerce(text_type(arg)) if res is Invalid else res


@coercer
def str_coerce(arg):
    '''Coerce to standard string type.

    `bytes` in Python 2.x and unicode (`str`) in Python 3.

    .. versionadded:: 1.7.0

    '''
    # TODO: Analyze if promote to global::
    #   str_coerce = unicode_coerce if _py3 else bytes_coerce
    from xoutil.eight import _py3
    return (unicode_coerce if _py3 else bytes_coerce)(arg)


@coercer
def ascii_coerce(arg):
    '''Coerce to string containing only ASCII characters.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    import unicodedata
    from xoutil.eight import text_type
    if not isinstance(arg, text_type):
        arg = unicode_coerce(arg)
    res = unicodedata.normalize('NFKD', arg).encode('ascii', 'ignore')
    return str_coerce(res)


@coercer
def ascii_set_coerce(arg):
    '''Coerce to string with only ASCII characters removing repetitions.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    return str().join(set(ascii_coerce(arg)))


@coercer
def lower_ascii_coerce(arg):
    '''Coerce to string containing only lower-case ASCII characters.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    return ascii_coerce(arg).lower()


@coercer
def lower_ascii_set_coerce(arg):
    '''Coerce to string with only lower-case ASCII chars removing repetitions.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    '''
    return str().join(set(lower_ascii_coerce(arg)))


def chars_coerce(arg):
    '''Convert to unicode characters.

    If `arg` is an integer between ``0`` and ``0x10ffff`` is converted
    assuming it as ordinal unicode code, else is converted with
    `unicode_coerce`:meth:.

    '''
    from xoutil.eight import _py3, integer_types as ints
    if isinstance(arg, ints) and 0 <= arg <= 0x10ffff:
        return (chr if _py3 else unichr)(arg)
    else:
        return unicode_coerce(arg)


from xoutil.eight import text_type as text


class text(text):
    '''Return a nice text representation of one object.

    text(obj='') -> text

    text(bytes_or_buffer[, encoding[, errors]]) -> text

    Create a new string object from the given object.  If `encoding` or
    `errors` is specified, then the object must expose a data buffer that will
    be decoded using the given encoding and error handler.  Otherwise, returns
    the result of object text representation.

    :param encoding: defaults to ``sys.getdefaultencoding()``.

    :param errors: defaults to 'strict'.

    Method join is improved, in order to receive any collection of objects,
    as variable number of arguments or as one iterable.

    '''
    def __new__(cls, obj='', *args, **kwargs):
        if not (args or kwargs):
            obj = unicode_coerce(obj)
        return super(text, cls).__new__(cls, obj, *args, **kwargs)

    def join(self, *args):
        '''S.join(variable_number_args or iterable) -> text

        Return a text which is the concatenation of the objects (converted to
        text) in argument items.  The separator between elements is `S`.

        See `chr_join`:meth: for other vertion of this functionality.

        '''
        return self._join(unicode_coerce, args)

    def chr_join(self, *args):
        '''S.chr_join(variable_number_args or iterable) -> text

        Return a text which is the concatenation of the objects (converted to
        text) in argument items.  The separator between elements is `S`.

        Difference with `join`:meth: is that integers between ``0`` and
        ``0x10ffff`` are converted to characters as unicode ordinal.

        '''
        return self._join(chars_coerce, args)

    def _join(self, coercer, args):
        '''Protected method to implement `join`:meth: and `chr_join`:meth:.'''
        from collections import Iterable
        if len(args) == 1 and isinstance(args[0], Iterable):
            args = args[0]
        return super(text, self).join(coercer(obj) for obj in args)
